#include "RdmaEndpoint.hpp"
int RdmaEndpoint::CONN_STATE_INITIALIZED = 1;
int RdmaEndpoint::CONN_STATE_RESOURCES_ALLOCATED = 2;
int RdmaEndpoint::CONN_STATE_CONNECTED = 3;
int RdmaEndpoint::CONN_STATE_CLOSED = 4;


RdmaEndpoint::RdmaEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                                       int recvQueueSize, int sendMsgSize, int recvMsgSize)
    : _cm_id(id), _completionQueue(completionQueue), _sendQueueSize(sendQueueSize),
      _recvQueueSize(recvQueueSize), _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize)
{
    _state = CONN_STATE_INITIALIZED;
    _sendBuffers = new boost::lockfree::queue<void*>(_sendMsgSize);
}
void RdmaEndpoint::createResources()
{
    if (_state != CONN_STATE_INITIALIZED)
    {
        std::cout << "RdmaEndpoint : createResource invalid satte" << std::endl;
    }

    _protectionDomain = ibv_alloc_pd(_cm_id->verbs);
    if (_protectionDomain == NULL)
    {
        std::cout << "RdmaEndpoint : ibv_alloc_pd failed " << std::endl;
        return;
    }

    struct ibv_qp_init_attr qp_init_attr;
    memset(&qp_init_attr, 0, sizeof(qp_init_attr));
    //This is used to set endpoint address with qp
    qp_init_attr.qp_context = (void *)this;
    // if not set 0, all work requests submitted to SQ will always generate a Work Completion
    qp_init_attr.sq_sig_all = 1;
    // completion queue can be shared or you can use distinct completion queues.
    qp_init_attr.send_cq = _completionQueue;
    qp_init_attr.recv_cq = _completionQueue;
    qp_init_attr.qp_type = IBV_QPT_RC;
    // increase if you want to keep more send work requests in the SQ.
    qp_init_attr.cap.max_send_wr = _sendQueueSize;
    // increase if you want to keep more receive work requests in the RQ.
    qp_init_attr.cap.max_recv_wr = _recvQueueSize;
    // increase if you allow send work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_send_sge = 1;
    // increase if you allow receive work requests to have multiple scatter gather entry (SGE).
    qp_init_attr.cap.max_recv_sge = 1;

    int ret = rdma_create_qp(_cm_id, _protectionDomain, &qp_init_attr);
    if (ret)
    {
        std::cout << "RdmaEndpoint : ibv_create_cq failed\n";
    }
    if (_cm_id->pd == NULL)
    {
        std::cout << "RdmaEndpoint : pd not set" << std::endl;
        _cm_id->pd = _protectionDomain;
    }
    _sendBuff = (char*)malloc(_sendMsgSize * _sendQueueSize);
    if (_sendBuff == NULL)
        std::cout << "RdmaEndpoint :  sendBuff malloc failed" << std::endl;
    _recvBuff = (char*)malloc(_recvMsgSize * _recvQueueSize);

    _sendMr = rdma_reg_write(_cm_id, reinterpret_cast<void *>(_sendBuff), _sendMsgSize * _sendQueueSize);
    if (_sendMr == NULL)
        std::cout << "RdmaEndpoint :  sendMr reg failed" << std::endl;

    if (_recvBuff == NULL)
        std::cout << "RdmaEndpoint :  recvBuff malloc failed" << std::endl;
    _recvMr = rdma_reg_read(_cm_id, reinterpret_cast<void *>(_recvBuff), _recvMsgSize * _recvQueueSize);
    if (_recvMr == NULL)
        std::cout << "RdmaEndpoint :  recvMr reg failed" << std::endl;

    for (int i = 0; i < _recvQueueSize; i++)
    {
        char *const location = _recvBuff + i * _recvMsgSize;
        rdma_post_recv(_cm_id, reinterpret_cast<void *>(location), reinterpret_cast<void *>(location),
                       _recvMsgSize, _recvMr);
    }
    
    for (int i = 0; i < _sendQueueSize; i++)
    {
        void* const location  = _sendBuff + i * _sendMsgSize;
        _sendBuffers->push(location);
    }

    _state = CONN_STATE_RESOURCES_ALLOCATED;
}


void RdmaEndpoint::processCmEvent(struct rdma_cm_event *event)
{
    std::cout << "RdmaEndpoint : Event " << rdma_event_str(event->event) << std::endl;
    if (event->event == RDMA_CM_EVENT_CONNECT_REQUEST)
    {
        std::cout << "RdmaEndpoint : Connect request";
    }
    else if (event->event == RDMA_CM_EVENT_ESTABLISHED)
    {
        if (_state != CONN_STATE_RESOURCES_ALLOCATED)
        {
            std::cout << "RdmaEndpoint : EstablishedEvent invalid state " << std::endl;
        }
        std::cout << "RdmaEndpoint : step 6 Connected" << std::endl;
        _state = CONN_STATE_CONNECTED;
    }
    else if (event->event == RDMA_CM_EVENT_DISCONNECTED)
    {
        std::cout << "RdmaEndpoint : step 7 disconnected" << std::endl;
        clientClose();
    }
}

void RdmaEndpoint::clientClose()
{
    if (_state != CONN_STATE_CONNECTED)
    {
        std::cout << "RdmaEndpoint :  clientClose invalid state" << std::endl;
        return;
    }
    std::cout<<"RdmaEndpoint : closing connection qp "<<_cm_id->qp->qp_num<<  std::endl;
    int ret;
    ret = rdma_disconnect(_cm_id);
    if (ret)
    {
        std::cout << "RdmaEndpoint : rdma_disconnect failed" << std::endl;
    }
    
    ret = rdma_dereg_mr(_sendMr);
    if (ret)
    {
        std::cout << "RdmaEndpoint : rdma_dereg_mr send failed" << std::endl;
    }
    free(_sendBuff);
    
    ret = rdma_dereg_mr(_recvMr);
    if (ret)
    {
        std::cout << "RdmaEndpoint : rdma_dereg_mr recv failed" << std::endl;
    }
    free(_recvBuff);
    
    rdma_destroy_qp(_cm_id);

    std::cout<<"des qp"<<std::endl;
   // ret = rdma_destroy_id(_cm_id);

    std::cout<<"des mr"<<std::endl;
    if (ret)
    {
        std::cout << "RdmaEndpoint : rdma_destroy_id failed" << std::endl;
    }
    _state = CONN_STATE_CLOSED;
    std::cout<<"closed"<<std::endl;
}
