#include "RdmaServerEndpointGroup.hpp"

int RdmaServerEndpointGroup::CONN_STATE_INITIALIZED = 0;
int RdmaServerEndpointGroup::CONN_STATE_BINDED = 3;
int RdmaServerEndpointGroup::CONN_STATE_CONNECTED = 4;
int RdmaServerEndpointGroup::CONN_STATE_CLOSED = 5;

RdmaServerEndpointGroup::RdmaServerEndpointGroup(int sendQueueSize, int recvQueueSize, int compQueueSize, int sendMsgSize, int recvMsgSize)
    : _sendQueueSize(sendQueueSize), _recvQueueSize(recvQueueSize), _compQueueSize(compQueueSize),
      _sendMsgSize(sendMsgSize), _recvMsgSize(recvMsgSize)
{
    _cmProcessor = new RdmaCmProcessor(this);
    _cm_id = _cmProcessor->createId();
    _qpRepEndpointMap = new std::unordered_map<uint32_t, RdmaReplicationEndpoint *>();
    _qpSalEndpointMap = new std::unordered_map<uint32_t, RdmaSalEndpoint *>();
    _salEps = new std::vector<RdmaSalEndpoint *>();
    _repEps = new std::vector<RdmaReplicationEndpoint *>();


    rocksdb::Options options;
    options.create_if_missing = true;
    // open a database with a name which corresponds to a file system directory
    rocksdb::Status status = rocksdb::DB::Open(options, "/tmp/testdb", &_db);
    std::cout << "Rocks started" << std::endl;
    if (!status.ok())
    {
        std::cout << status.ToString() << std::endl;
        exit(1);
    }
}
void RdmaServerEndpointGroup::setExecutor(Executor* ex)
{
    _executor = ex;
}
void RdmaServerEndpointGroup::startCmProcessor(bool newThread)
{
    _cmProcessor->start(newThread);
}

void RdmaServerEndpointGroup::bind(const char *ip, const char *port, int backlog)
{
    int ret;
    std::cout << "RdmaServerEndpointGroup : Step 2 bind_addr" << std::endl;
    struct addrinfo *addr;
    ret = getaddrinfo(ip, port, NULL, &addr);
    if (ret)
    {
        std::cout << "RdmaServerEndpointGroup : get_addr_info failed" << std::endl;
    }
    ret = rdma_bind_addr(_cm_id, addr->ai_addr);
    if (ret)
    {
        std::cout << "RdmaServerEndpointGroup : rdma_bind_addr failed" << std::endl;
        return;
    }

    std::cout << "RdmaServerEndpointGroup : rdma_bind_addr successful\n";
    ret = rdma_listen(_cm_id, backlog);
    if (ret)
    {
        std::cout << "RdmaServerEndpointGroup : rdma_listen failed" << std::endl;
        return;
    }
}

struct ibv_cq *RdmaServerEndpointGroup::createSalCq(struct rdma_cm_id *id)
{
    if (_salCqProcessor == NULL)
    {
        std::cout << "RdmaServerEndpointGroup : step 5 create salcq processor" << std::endl;
        _salCqProcessor = new RdmaSalCqProcessor(_executor, _cm_id->verbs, _compQueueSize);
        _salCqProcessor->start();
    }
    return _salCqProcessor->getCq();
}

struct ibv_cq *RdmaServerEndpointGroup::createRepCq(struct rdma_cm_id *id)
{
    if (_repCqProcessor == NULL)
    {
        std::cout << "RdmaServerEndpointGroup : step 5 create repcq processor" << std::endl;
        _repCqProcessor = new RdmaRepCqProcessor(_executor, _cm_id->verbs, _compQueueSize);
        _repCqProcessor->start();
    }
    return _repCqProcessor->getCq();
}
void RdmaServerEndpointGroup::createEpCmEvent(struct rdma_cm_event *event)
{
    std::cout << "RdmaServerEndpointGroup : step 4 Got Connect Request Server Endpoint" << std::endl; 
    /* create and add to vectors for replication and invalidation processing*/
    const char *connData = reinterpret_cast<const char *>(event->param.conn.private_data);
    if (strcmp(connData, "sal") == 0)
    {
        std::cout << "sal" << std::endl;
        RdmaSalEndpoint *endpoint = new RdmaSalEndpoint(event->id, createSalCq(event->id), _sendQueueSize, _recvQueueSize,
                                                        _sendMsgSize, _recvMsgSize,_db);
        event->id->context = (void *)endpoint;
        endpoint->createResources();
        std::unique_lock lock(_salMutex);
        _salEps->push_back(endpoint);
        std::cout<<"qp num"<<event->id->qp->qp_num<<" "<<_qpSalEndpointMap->size();
        _qpSalEndpointMap->emplace(event->id->qp->qp_num, endpoint);
        std::cout<<_qpSalEndpointMap->size()<<std::endl;
    }
    else
    {
        RdmaReplicationEndpoint *endpoint = new RdmaReplicationEndpoint(event->id, createRepCq(event->id), _sendQueueSize, _recvQueueSize,
                                                                        _sendMsgSize, _recvMsgSize,_db);
        event->id->context = (void *)endpoint;
        endpoint->createResources();
        std::unique_lock lock(_repMutex);
        _repEps->push_back(endpoint);
        _qpRepEndpointMap->emplace(event->id->qp->qp_num, endpoint);
    }
    rdma_accept(event->id, NULL);
}

void RdmaServerEndpointGroup::processCmEvent(struct rdma_cm_event *event)
{
    std::cout << "RdmaServerEndpointGroup : event " << rdma_event_str(event->event) << " id " << event->id << " " << std::endl;
    /*
    * Connect request came on listener ie endpointgroup
    * in this listen_id contains cm_id of listener and id contains the new id of connection request
    */
    if (event->event == RDMA_CM_EVENT_CONNECT_REQUEST && event->listen_id != NULL)
    {
        createEpCmEvent(event);
    }
    /*
    * Event came for listener ie endpointgroup
    */
    else if (event->id != NULL && _cm_id == event->id)
    {
        if (event->event == RDMA_CM_EVENT_DISCONNECTED)
        {
            std::cout << "RdmaServerEndpointGroup : Disconnect Server Endpoint" << std::endl;
            close();
        }
        else
        {
            std::cout << "RdmaServerEndpointGroup : unknown Event for listener" << rdma_event_str(event->event) << std::endl;
        }
    }
    /*
    * Event came for a endpoint
    */
    else if (event->id != NULL && event->id->context != NULL)
    {
        RdmaEndpoint *ep = ((RdmaEndpoint *)event->id->context);
        ep->processCmEvent(event);
        if (event->event == RDMA_CM_EVENT_DISCONNECTED)
        {
            {
                std::unique_lock lock(_repMutex);
                auto it = std::find(_repEps->begin(), _repEps->end(), (RdmaReplicationEndpoint *)ep);
                if (it != _repEps->end())
                {
                    _repEps->erase(it);
                    delete ((RdmaReplicationEndpoint *)event->id->context);
                }
            }

            {
                std::unique_lock lock(_salMutex);
                auto it = std::find(_salEps->begin(), _salEps->end(), (RdmaSalEndpoint *)ep);
                if (it != _salEps->end())
                {
                    _salEps->erase(it);
                    delete ((RdmaSalEndpoint *)event->id->context);
                }
            }
        }

        /*if (event->event == RDMA_CM_EVENT_TIMEWAIT_EXIT)
        {
            std::cout<<"des"<<std::endl;
            rdma_destroy_id(event->id);
            std::cout<<"des"<<std::endl;
        }
        */
    }
    else
    {
        std::cout << "RdmaServerEndpointGroup :  Not able to procces CM EVent" << rdma_event_str(event->event) << event->id << " " << event->listen_id << std::endl;
    }
}

void RdmaServerEndpointGroup::clientClose()
{
    _cmProcessor->close();
    delete _cmProcessor;
    _salCqProcessor->close();
    _repCqProcessor->close();
    delete _salCqProcessor;
    delete _repCqProcessor;
    /* for(int i = 0 ;i < _endPoints->size();i++)
    {
        _endPoints->at(i)->close();
    }
    delete _endPoints;
    */
    rdma_disconnect(_cm_id);
    rdma_destroy_id(_cm_id);
}

void RdmaServerEndpointGroup::close()
{
}

